#   Google Spreadsheet Reports for OpenERP
#   Copyright 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Affero Public License version 3, as published 
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranties of 
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#   PURPOSE.  See the GNU Affero Public License for more details.
#
#   You should have received a copy of the GNU Affero Public License along 
#   with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gdata.docs.service import DocsService
from gdata.spreadsheet.service import SpreadsheetsService
from gdata.spreadsheet.text_db import DatabaseClient
import httplib

class Connection(object):
    """
    Google GData connection utility class
    """
    
    def __init__(self, email, password):
        """
        """
        self.email = email
        self.password = password
        # Uncomment next line to enable debugging HTTP messages
        #httplib.HTTPConnection.debuglevel = 1

class DocConnection(Connection):
    """
    Connection for talking with Google Docs.
    """
    def __init__(self, email, password):
        super(DocConnection, self).__init__(email, password)
        self.client = DocsService()
        self.client.ssl = True
        self.client.ClientLogin(self.email, self.password)

    def get_docs(self):
        documents_feed = self.client.GetDocumentListFeed()
        return documents_feed.entry

    def create_doc(self, title, label="spreadsheet"):
        doc = self.client.Upload(title=title, label=label)
        return doc



class SpreadsheetConnection(Connection):
    """
    Connection for Google Spreadsheet.
    """
    def __init__(self, email, password):
        super(SpreadsheetConnection, self).__init__(email, password)
        self.client = SpreadsheetsService()
        self.client.ssl = True
        self.client.email = self.email
        self.client.password = self.password
        self.client.source = "Canonical_Gdata"
        self.client.ProgrammaticLogin()

    def create_spreadsheet(self, name):
        """
        There is no create spreadsheet method, need to use upload document.
        However, that doesn't use SSL, so implementing raw POST instead.
        """
        # Login to the Docs service
        doc = DocConnection(self.email, self.password)

        data = u'<?xml version=\'1.0\' encoding=\'UTF-8\'?>\n<ns0:entry xmlns:ns0="http://www.w3.org/2005/Atom"><ns0:category label="spreadsheet" scheme="http://schemas.google.com/g/2005#kind" term="http://schemas.google.com/docs/2007#spreadsheet" /><ns0:title>' + name + u'</ns0:title></ns0:entry>'

        # Create the spreadsheet
        uri = '/feeds/documents/private/full'
        entry = doc.client.Post(data, uri)
        
        # Return the created spreadsheet as an object
        return self.get_spreadsheet(name)

    def get_document_id(self, name):
        # Get the spreadsheets
        sheets = self.client.GetSpreadsheetsFeed()
        spreadsheet = None
        for s in sheets.entry:
            if s.title.text == name:
                spreadsheet = s.id.text.rsplit('/', 1)[1]

        return spreadsheet

    def get_spreadsheets(self):
        sheets = self.client.GetSpreadsheetsFeed()
        return sheets.entry

    def get_spreadsheet(self, name):
        sheets = self.get_spreadsheets()
        named_sheet = None
        for sheet in sheets:
            if sheet.title.text == name:
                named_sheet = sheet
                break
        return named_sheet

    def get_worksheets(ss_key):
        feed = self.client.GetWorksheetsFeed(ss_key)
        return feed.entry
            
    def get_worksheet(self, ss_key, name):
        feed = self.client.GetWorksheetsFeed(ss_key)
        named_sheet = None
        for w in feed.entry:
            if w.title.text == name:
                named_sheet = w
                break

        return named_sheet

    def get_worksheet_id(self, ss_key, name):
        sheets = self.get_worksheets(ss_key)
        wrk_id = None
        for w in sheets:
            if w.title.text == name:
                wrk_id = w.id.text.rsplit('/', 1)[1]

        return wrk_id

    def create_worksheet(self, ss_key, name, cols=20, rows=1):
        sheet = self.client.AddWorksheet(title=name, row_count=rows, col_count=cols, key=ss_key)
        return sheet
    
    def delete_worksheet(self, worksheet):
        return self.client.DeleteWorksheet(worksheet)

    def insert_row(self, ss_key, work_id, row):
        entry = self.client.InsertRow(row, ss_key, work_id)
        return entry

    def get_key(self, entry):
        return entry.id.text.rsplit('/', 1)[1]

    def create_header_row(self, ss_key, ws_key, fields):
        for i in range(0, len(fields)):
            self.client.UpdateCell(1, i+1, fields[i], ss_key, ws_key)

    def delete_rows(self, ss_key, ws_key):
        rows = self.client.GetListFeed(ss_key, ws_key)
        if len(rows.entry) > 0:
            for r in rows.entry:
                self.client.DeleteRow(r)

    def update_cell(self, ss_key, ws_key, row, col, value):
        return self.client.UpdateCell(row, col, value, ss_key, ws_key)
        

class DatabaseConnection(Connection):
    """
    Connection for talking to Google Spreadsheets as though they are databases.
    A worksheet (or part of it) can represent a table.
    """
    def __init__(self, email, password):
        super(DatabaseConnection, self).__init__(email, password)
        self.client = DatabaseClient()
        self.client.ssl = True
        self.client.SetCredentials(self.email, self.password)

    def create_database(self, name):
        db = self.client.CreateDatabase(name)
        return db

    def get_databases(self, name=None):
        dbs = self.client.GetDatabases(name=name)
        return dbs

    def create_table(self, database, name, fields):
        table = database.CreateTable(name, fields)
        return table

    def get_tables(self, db, name=None):
        tables = db.GetTables(name=name)
        return tables
       
    def get_row(self, table, row_no):
        row = table.GetRecord(row_number=row_no)
        if row:
            return row.content
        else:
            return

    def get_rows(self, table, start_row, end_row):
        rows = table.GetRecords(start_row, end_row)
        return rows

    def get_fields(self, table):
        """
        This method always returns 'None' due to a bug in LookupFields()
        """
        fields = table.LookupFields()
        return fields

    def truncate_table(self, table):
        rows = self.get_rows(table, 1,100)
        count = len(rows)
        while count > 0:            
            for r in rows:
                r.Delete()
            rows = self.get_rows(table, 1,100)
            count = len(rows)

    def add_record(self, table, record):
        rec = table.AddRecord(record)
        return rec

