#   Google Spreadsheet Reports for OpenERP
#   Copyright 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Affero Public License version 3, as published 
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranties of 
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#   PURPOSE.  See the GNU Affero Public License for more details.
#
#   You should have received a copy of the GNU Affero Public License along 
#   with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import time, datetime
import re
import base64

from osv import fields,osv,orm
from tools.translate import _
from google.connection import DatabaseConnection
from google.connection import SpreadsheetConnection
from tools.misc import email_send
from tools.config import config

def str2tuple(s):
    return eval('tuple(%s)' % (s or ''))

class report_gdata(osv.osv):
    _name ='report.gdata.report'
    _columns = {
        'name': fields.char('Report name', size=200, required=True, help='Descriptive name for the report.'),
        'google_email': fields.char('Google Email', size=64, required=True, help='Email address used to logon to the Google account.'),
        'google_password': fields.char('Google password', size=64, required=True, help='Password used to logon to the Google account.'),
        'spreadsheet': fields.char("Spreadsheet name", size=80, required=True, help='The name of the Google spreadsheet that is to be updated. This will be created if it does not exist.'),
        'worksheet': fields.char('Worksheet name', size=20, required=True, help='The name of the worksheet on the Google spreadsheet that is to be updated. If this exists, all records on this worksheet will be removed when the report is run.'),
        'type': fields.selection([('sql','SQL Query'),('function','Python Function')], string="Type"),
        'model': fields.char('Object', size=64, help="Name of object whose function will be called when this scheduler will run. e.g. 'res.partner'"),
        'function': fields.char('Function', size=64, help="Name of the method to be called on the object when this scheduler is executed. It must return a list of rows."),
        'query': fields.text('SQL query/Arguments', required=True, help='The database query that will retrieve the fields for the report, or the arguments (for the Python function).'),
        'interval': fields.selection([('unscheduled','Unscheduled'),('daily','Daily'),('weekly','Weekly'),('monthly','Monthly')], 'Interval', required=True, ),
        'week_day': fields.selection([('1','Monday'),('2','Tuesday'),('3','Wednesday'),('4','Thursday'),('5','Friday'),('6','Saturday'),('0','Sunday')], 'Day'),
        'month_day': fields.integer("Day of Month"),
        'at_time': fields.time("At Time"),
        'next_run': fields.datetime('Next run'),
        'log': fields.text('Log'),
        'notify': fields.selection([('none','None'),('users','Users')], 'Email Notification', required=True),
        'user_ids': fields.many2many('res.users', 'report_gdata_users_rel', 'report_id','user_id', 'Users'),
         }

    _defaults = {
        'type': lambda *a: 'sql',
        'interval': lambda *a: 'unscheduled',
        'month_day': lambda *a: 1,
        'week_day': lambda *a: '1',
        }

    def run(self, cr, uid, ids, manual=False, context=None):
        """
        Run the report by running the SQL query and creating/updating the spreadsheet with the data.
        """
        # Get the report record(s)
        self.gdata_ss = None
        for rep in self.browse(cr, uid, ids, context):
            try:
                log = 'Starting report %s\n' % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                spreadsheet = None

                # Create a spreadsheet connection
                if not self.gdata_ss:
                    self.gdata_ss = SpreadsheetConnection(rep.google_email, base64.b64decode(rep.google_password))
                 
                spreadsheet = self._get_spreadsheet(rep)
                log += 'Spreadsheet link: %s\n' % spreadsheet.GetAlternateLink().href
                
                # Run the report depending on the report type
                if rep.type == None or rep.type=='sql':
                    # Run report as an SQL query
                    log += self.run_sql(cr, uid, rep, spreadsheet, manual=manual, context=context)
                else:
                    # Run Python method
                    log += self.run_function(cr, uid, rep, spreadsheet, manual=manual, context=context)

                # Set the next run by saving the record
                log += "Report complete: %s\n" % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                rec_upd = {'log': log}
                if not manual:
                    next_run = self._inc_next_run(rep.id, rep.interval, rep.next_run)
                    rec_upd['next_run'] = next_run
                self.write(cr, uid, [rep.id], rec_upd, context)

                # Send the Email notifications
                if rep.notify == 'users':
                    self.send_emails(rep, spreadsheet.GetAlternateLink().href)

            except Exception as e:
                # Email error notification
                if rep.notify:
                    if spreadsheet:
                        link = spreadsheet.GetAlternateLink().href
                    else:
                        link = 'Unknown'
                    self.send_error_emails(rep, link, str(e))

                # Save the error details in the log
                log += 'Error: %s' % str(e)
                rec_upd = {'log': log}
                self.write(cr, uid, [rep.id], rec_upd, context)

    def run_sql(self, cr, uid, rep, spreadsheet, manual=False, context=None):
        """
        Run a SQL Query report.
        """
        log = ''
        
        # Run the SQL query
        cr.execute(rep.query)
        recs = cr.dictfetchall()
        log += 'SQL query returned %d rows\n' % len(recs)

        # Get the field names from the query and sanitise them
        if len(recs) > 0:
            fields = [self._clean_name(x) for x in recs[0].keys()]
        else:
            fields = []

        # Get/create the table
        worksheet = self._get_worksheet(spreadsheet, rep, fields)

        # Add the records to the spreadsheet
        ss_key = self.gdata_ss.get_key(spreadsheet)
        ws_key = self.gdata_ss.get_key(worksheet)
        for rec in recs:
            # Set up a Google Doc friendly record dictionary
            record = {}
            for key, value in rec.items():
                k = self._clean_name(key)
                record[k] = str(value)

            # Add the record to the spreadsheet
            self.gdata_ss.insert_row(ss_key, ws_key, record)

        return log

    def run_function(self, cr, uid, rep, spreadsheet, manual=False, context=None):
        """
        Run the report as a Python function.
        This similar to how ir.cron works, but the method must return a list of rows.
        These will be used to update the spreadsheet.
        """
        log = "Running %s.%s\n" % (rep.model, rep.function)

        # Attempt to run the Python method
        args = str2tuple(rep.query)
        m = self.pool.get(rep.model)
        if m and hasattr(m, rep.function):
            f = getattr(m, rep.function)
            try:
                # Call the method to return a list of rows
                rows = f(cr, uid, *args)

                if rows:
                    # Delete/create the worksheet
                    worksheet = self._create_worksheet(spreadsheet,rep, cols=30, rows=len(rows))

                    # Get the spreadsheet/worksheet keys
                    ss_key = self.gdata_ss.get_key(spreadsheet)
                    ws_key = self.gdata_ss.get_key(worksheet)

                    # Update the worksheet with the data
                    rnum = 1
                    for row in rows:
                        cnum = 1
                        for r in row:
                            self.gdata_ss.update_cell(ss_key, ws_key, rnum, cnum, str(r))
                            cnum += 1
                        rnum += 1
                    
            except Exception, e:
                cr.rollback()
                raise
                
        else:
            print "Not found", m, rep.model, rep.function, getattr(m, rep.function)
            raise orm.except_orm("Error running report '%s'" % rep.name,
                "Cannot find method '%s' for object '%s' in report '%s'" % (rep.function, rep.model, rep.name))
        return log


    def send_emails(self, rep, link):
        # Send the email notification
        subject = 'Report updated: %s' % rep.name
        message = 'The report %s has been updated.\n View the updated spreadsheet:\n %s' % (rep.name, link)
        emails = [user.address_id.email for user in rep.user_ids if user.address_id and user.address_id.email]

        email_send(config['email_from'], emails, subject, message)

    def send_error_emails(self, rep, link, error):
        # Send the error notification
        subject = 'Error updating report: %s' % rep.name
        message = 'The following error was encountered when running report %s.\n View the updated spreadsheet:\n %s\n\n%s' % (rep.name, link, error)
        emails = [user.address_id.email for user in rep.user_ids if user.address_id and user.address_id.email]

        email_send(config['email_from'], emails, subject, message)

    def _get_spreadsheet(self, rep):
        spreadsheets = self.gdata_ss.get_spreadsheets()
        spreadsheet = None

        if len(spreadsheets) == 0:
            # Create the spreadsheet
            spreadsheet = self.gdata_ss.create_spreadsheet(rep.spreadsheet)
        else:
            # Check that the sheet name matches, otherwise create new database
            for ssheet in spreadsheets:
                if ssheet.title.text == rep.spreadsheet:
                    spreadsheet = ssheet
                    break
            if not spreadsheet:
                spreadsheet = self.gdata_ss.create_spreadsheet(rep.spreadsheet)

        return spreadsheet

    def _get_worksheet(self, spreadsheet, rep, fields):
        ss_key = self.gdata_ss.get_key(spreadsheet)
        worksheet = self.gdata_ss.get_worksheet(ss_key, rep.worksheet)

        if not worksheet:
            # Create the worksheet
            worksheet = self.gdata_ss.create_worksheet(ss_key, rep.worksheet, len(fields))
            ws_id = self.gdata_ss.get_key(worksheet)

            self.gdata_ss.create_header_row(ss_key, ws_id, fields)
        else:
            # Remove existing records in the table (optionally?)
            ws_key = self.gdata_ss.get_key(worksheet)
            self.gdata_ss.delete_rows(ss_key, ws_key)

        return worksheet

    def _create_worksheet(self, spreadsheet, rep, cols, rows):
        ss_key = self.gdata_ss.get_key(spreadsheet)
        worksheet = self.gdata_ss.get_worksheet(ss_key, rep.worksheet)

        if worksheet:
            # Delete the worksheet
            self.gdata_ss.delete_worksheet(worksheet)
        worksheet = self.gdata_ss.create_worksheet(ss_key, rep.worksheet, cols=cols, rows=rows)        
        return worksheet

    def _clean_name(self, name):
        """
        Google Doc spreadsheet tables work best with alphanumeric fieldnames, lowercase
        """
        return re.sub(r'[^a-zA-Z0-9]', '', name).lower()

    def run_report_cron(self, cr, uid):
        # Get the reports that need running
        now = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
        domain = [('interval','!=','unscheduled'),('next_run','<=',now)]
        ids = self.search(cr, uid, domain)
        
        # Run each of the reports
        self.run(cr, uid, ids)

    def run_report_click(self, cr, uid, ids, context=None):
        self.run(cr, uid, ids, True, context)
        return True

    def create(self, cr, uid, vals, context=None):
        record = vals.copy()
        if record.has_key('google_password'):
            record['google_password'] = base64.b64encode(vals['google_password'])
        return super(report_gdata, self).create(cr, uid, record, context)

    def write(self, cr, uid, ids, vals, context=None):
        # Get the existing record
        for record in  self.read(cr, uid, ids, [], context):
            # Update the dictionary to get the full record
            record.update(vals)
            rec_id = record.pop('id')
            
            # Calculate the next run date if it is not supplied
            if not vals.has_key('next_run'):
                record = self._calc_next_run(record)

            # Hash the password, if it has been changed
            if vals.has_key('google_password'):
                record['google_password'] = base64.b64encode(vals['google_password'])

            # Update the record
            super(report_gdata, self).write(cr, uid, rec_id, record, context)
        return True

    def _calc_next_run(self, record):
        """
        Calculate the next run date/time.
        """
        if record['interval'] == 'unscheduled':
            return record
        elif record['interval'] == 'weekly':
            # Use the weekday to calculate the run date for this week
            day = int(record['week_day']) - 1
            date = datetime.date.today() - datetime.timedelta(days=datetime.date.today().weekday()) + datetime.timedelta(days=day)
            if date < datetime.date.today():
                # Run has passed so plan for next week
                date = date + datetime.timedelta(days=7)
        elif record['interval'] == 'monthly':
            # Use the month day to calculate the date of the next run
            day = int(record['month_day'])
            (year,month) = datetime.date.today().timetuple()[:2]
            date = datetime.date(year,month,day)
            if date < datetime.date.today():
                # Run date has passed so plan for next month
                date = self._add_months(date, 1)
        elif record['interval'] == 'daily':
            # Use the time to schedule the run for each day
            date = datetime.date.today()

        record['next_run'] = '%s %s' % (date.strftime('%Y-%m-%d'), record['at_time'])
        return record

    def _inc_next_run(self, rep_id, interval, next_run):
        """
        Update the report's next run date after it has been run.
        """
        date = False
        if interval == 'weekly':
            next_date = datetime.datetime.strptime(next_run, '%Y-%m-%d %H:%M:%S') + datetime.timedelta(days=7)
            date = next_date.strftime('%Y-%m-%d %H:%M:%S')
        elif interval == 'monthly':
            next_date = self._add_months(next_run, 1)
            date = next_date.strftime('%Y-%m-%d %H:%M:%S')
        elif interval == 'daily':
            next_date = datetime.datetime.strptime(next_run, '%Y-%m-%d %H:%M:%S') + datetime.timedelta(days=1)
            date = next_date.strftime('%Y-%m-%d %H:%M:%S')
            
        return date

    def _add_months(self, date, months=1):
        """
        Adds a number of months to a date. The date can be a string or a datetime field.
        """
        if isinstance(date, basestring):
            this_date = datetime.datetime.strptime(date, '%Y-%m-%d %H:%M:%S')
        else:
            this_date = date
        (year,month,day,hours,mins,secs) = this_date.timetuple()[:6]
        new_month = month + 1
        
        if new_month > 12:
            year += 1
            new_month = 1

        next_date = datetime.datetime(year, new_month, day, hours, mins, secs)
        return next_date



report_gdata()
